/*=============================================================================
XMOTO
Copyright (C) 2005 Rasmus Neckelmann (neckelmann@gmail.com)

This file is part of XMOTO.

XMOTO is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

XMOTO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with XMOTO; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
=============================================================================*/

/* 
 *  In-game rendering
 */
#include "VXml.h"
#include "VFileIO.h"
#include "MotoGame.h"
#include "Renderer.h"

namespace vapp {
  
  /*===========================================================================
  Called to prepare renderer for new level
  ===========================================================================*/
  void GameRenderer::prepareForNewLevel(void) {
    m_fCurrentHorizontalScrollShift = 0.0f;
    m_fDesiredHorizontalScrollShift = 0.0f;
    
    m_fNextParticleUpdate = 0.0f;
  }

  /*===========================================================================
  Called when we don't want to play the level anymore
  ===========================================================================*/
  void GameRenderer::unprepareForNewLevel(void) {
    m_pInGameStats->showWindow(false);
    
    /* Free any particles left */
    for(int i=0;i<m_Particles.size();i++)
      delete m_Particles[i];    
    m_Particles.clear();
  }  

  /*===========================================================================
  Minimap rendering
  ===========================================================================*/
  #define MINIMAPZOOM 5.0f
  #define MINIMAPALPHA 128
  #define MINIVERTEX(Px,Py) \
    getParent()->glVertex(x + nWidth/2 + (float)(Px + m_Scroll.x)*MINIMAPZOOM, \
                          y + nHeight/2 - (float)(Py + m_Scroll.y)*MINIMAPZOOM);    

  void GameRenderer::renderMiniMap(int x,int y,int nWidth,int nHeight) {
    getParent()->drawBox(Vector2f(x,y),Vector2f(x+nWidth,y+nHeight),1,
                         MAKE_COLOR(0,0,0,MINIMAPALPHA),MAKE_COLOR(255,255,255,MINIMAPALPHA));
    getParent()->scissorGraphics(x+1,y+1,nWidth-2,nHeight-2);
    glEnable(GL_SCISSOR_TEST);
        
    MotoGame *pGame = getGameObject();
    std::vector<ConvexBlock *> &Blocks = pGame->getBlocks();

    for(int i=0;i<Blocks.size();i++) {
      Vector2f Center;

      if(Blocks[i]->pSrcBlock) {
        if(Blocks[i]->pSrcBlock->bBackground) continue;
        Center = Vector2f(Blocks[i]->pSrcBlock->fPosX,Blocks[i]->pSrcBlock->fPosY);
      }

      glBegin(GL_POLYGON);
      glColor3f(0.5,0.5,0.5);
      for(int j=0;j<Blocks[i]->Vertices.size();j++) {
        Vector2f P = Center + Blocks[i]->Vertices[j]->P;
        MINIVERTEX(P.x,P.y);                  
      }
      glEnd();
    }    
    
    getParent()->drawCircle(Vector2f(x + nWidth/2 + (float)(pGame->getBikeState()->CenterP.x + m_Scroll.x)*MINIMAPZOOM,
                                     y + nHeight/2 - (float)(pGame->getBikeState()->CenterP.y + m_Scroll.y)*MINIMAPZOOM),
                            3,0,MAKE_COLOR(255,255,255,255),0);
                            
    for(int i=0;i<pGame->getEntities().size();i++) {
      if(pGame->getEntities()[i]->Type == ET_ENDOFLEVEL) {
        getParent()->drawCircle(Vector2f(x + nWidth/2 + (float)(pGame->getEntities()[i]->Pos.x + m_Scroll.x)*MINIMAPZOOM,
                                        y + nHeight/2 - (float)(pGame->getEntities()[i]->Pos.y + m_Scroll.y)*MINIMAPZOOM),
                                3,0,MAKE_COLOR(255,0,255,255),0);
        
      }
      else if(pGame->getEntities()[i]->Type == ET_STRAWBERRY) {
        getParent()->drawCircle(Vector2f(x + nWidth/2 + (float)(pGame->getEntities()[i]->Pos.x + m_Scroll.x)*MINIMAPZOOM,
                                        y + nHeight/2 - (float)(pGame->getEntities()[i]->Pos.y + m_Scroll.y)*MINIMAPZOOM),
                                3,0,MAKE_COLOR(255,0,0,255),0);
        
      }
    }
        
    glDisable(GL_SCISSOR_TEST);
    getParent()->scissorGraphics(0,0,getParent()->getDispWidth(),getParent()->getDispHeight());
  }

  /*===========================================================================
  Main rendering function
  ===========================================================================*/
  void GameRenderer::render(void) {  
    /* Update time */    
    m_pInGameStats->showWindow(true);
    m_pPlayTime->setCaption(getParent()->formatTime(getGameObject()->getTime()));
  
    /* Prepare for rendering frame */
    _UpdateAnimations();
    
    if(getGameObject()->getTime() > m_fNextParticleUpdate) {
      _UpdateParticles(0.025f);
      m_fNextParticleUpdate = getGameObject()->getTime() + 0.025f; 
    }
    
    m_Scroll = -getGameObject()->getBikeState()->CenterP; /* Determine scroll */    
    m_fZoom = 60.0f;    
    
    /* Driving direction? */
    if(getGameObject()->getBikeState()->Dir == DD_RIGHT) {
      m_fDesiredHorizontalScrollShift = -4;
    }
    else if(getGameObject()->getBikeState()->Dir == DD_LEFT) {
      m_fDesiredHorizontalScrollShift = 4;
    }
    
    if(m_fDesiredHorizontalScrollShift != m_fCurrentHorizontalScrollShift) {
      float d = m_fDesiredHorizontalScrollShift - m_fCurrentHorizontalScrollShift;
      if(fabs(d)<0.25f) {
        m_fCurrentHorizontalScrollShift = m_fDesiredHorizontalScrollShift;
      }
      else if(d < 0.0f) {
        m_fCurrentHorizontalScrollShift -= 0.1f;
      }
      else if(d > 0.0f) {
        m_fCurrentHorizontalScrollShift += 0.1f;
      }
    }

    m_Scroll += Vector2f(m_fCurrentHorizontalScrollShift,0.0f);
  
    /* Start by rendering the background ... */
    _RenderBackground();
    
    /* ... then render background sprites ... */
    _RenderSprites(false,true);
        
    /* ... covered by blocks ... */
    _RenderBlocks();
    
    /* ... the entities ... */
    _RenderEntities();

    /* ... followed by the bike ... */
    _RenderBike();
    
    /* Render particles */
    _RenderParticles();
    
    /* ... and finally the foreground sprites! */
    _RenderSprites(true,false);
        
    /* And then the game messages */
    _RenderGameMessages();
            
    /* Hmm, in debug-mode we'd also like to see the zones and stuff */
    if(isDebug()) {    
      /* Render debug info */
      _RenderDebugInfo();

      /* Zones */
      for(int k=0;k<getGameObject()->getLevelSrc()->getZoneList().size();k++) {
        LevelZone *pZone = getGameObject()->getLevelSrc()->getZoneList()[k];
        
        for(int u=0;u<pZone->Prims.size();u++) {
          if(pZone->Prims[u]->Type == LZPT_BOX) {
            glBegin(GL_LINE_LOOP);
            glColor3f(0,1,0);
            getParent()->glVertex( getParent()->getDispWidth()/2 + (pZone->Prims[u]->fLeft + m_Scroll.x)*60.0f,
                      getParent()->getDispHeight()/2 - (pZone->Prims[u]->fTop + m_Scroll.y)*60.0f );
            getParent()->glVertex( getParent()->getDispWidth()/2 + (pZone->Prims[u]->fRight + m_Scroll.x)*60.0f,
                      getParent()->getDispHeight()/2 - (pZone->Prims[u]->fTop + m_Scroll.y)*60.0f );
            getParent()->glVertex( getParent()->getDispWidth()/2 + (pZone->Prims[u]->fRight + m_Scroll.x)*60.0f,
                      getParent()->getDispHeight()/2 - (pZone->Prims[u]->fBottom + m_Scroll.y)*60.0f );
            getParent()->glVertex( getParent()->getDispWidth()/2 + (pZone->Prims[u]->fLeft + m_Scroll.x)*60.0f,
                      getParent()->getDispHeight()/2 - (pZone->Prims[u]->fBottom + m_Scroll.y)*60.0f );
            glEnd();
            
            Vector2f TP(getParent()->getDispWidth()/2 + (pZone->Prims[u]->fLeft + m_Scroll.x)*60.0f,
                      getParent()->getDispHeight()/2 - (pZone->Prims[u]->fTop + m_Scroll.y)*60.0f);
            float th=12,tw=pZone->ID.length()*8;
            if(TP.x < 0) TP.x=0;
            if(TP.y < 0) TP.y=0;
            if(TP.x > getParent()->getDispWidth()-pZone->ID.length()*8) TP.x=getParent()->getDispWidth()-pZone->ID.length()*8;
            if(TP.y > getParent()->getDispHeight()-12) TP.y=getParent()->getDispHeight()-12;
            if(TP.x > getParent()->getDispWidth()/2 + (pZone->Prims[u]->fRight + m_Scroll.x)*60.0f) continue;
            if(TP.y > getParent()->getDispHeight()/2 - (pZone->Prims[u]->fBottom + m_Scroll.y)*60.0f) continue;
            if(TP.x + tw < getParent()->getDispWidth()/2 + (pZone->Prims[u]->fLeft + m_Scroll.x)*60.0f) continue;
            if(TP.y + th < getParent()->getDispHeight()/2 - (pZone->Prims[u]->fTop + m_Scroll.y)*60.0f) continue;
            getParent()->drawText(TP,pZone->ID,0,MAKE_COLOR(0,255,0,255),true);               
          }
        }
      }
    }        
  }
  
  /*===========================================================================
  Game message rendering
  ===========================================================================*/
  void GameRenderer::_RenderGameMessages(void) {
    MotoGame *pGame = getGameObject();

    /* Arrow messages */
    ArrowPointer *pArrow = &pGame->getArrowPointer();
    if(pArrow->nArrowPointerMode != 0) {
      Vector2f C;
      if(pArrow->nArrowPointerMode == 1) {          
        C=Vector2f(getParent()->getDispWidth()/2 + (float)(pArrow->ArrowPointerPos.x + m_Scroll.x)*m_fZoom,
                  getParent()->getDispHeight()/2 - (float)(pArrow->ArrowPointerPos.y + m_Scroll.y)*m_fZoom);      
      }
      else if(pArrow->nArrowPointerMode == 2) {          
        C=pArrow->ArrowPointerPos;
      }
      Vector2f p1,p2,p3,p4;
      p1 = Vector2f(1,0); p1.rotateXY(pArrow->fArrowPointerAngle);
      p2 = Vector2f(1,0); p2.rotateXY(90+pArrow->fArrowPointerAngle);
      p3 = Vector2f(1,0); p3.rotateXY(180+pArrow->fArrowPointerAngle);
      p4 = Vector2f(1,0); p4.rotateXY(270+pArrow->fArrowPointerAngle);

      p1 = p1 * 50.0f;
      p2 = p2 * 50.0f;
      p3 = p3 * 50.0f;
      p4 = p4 * 50.0f;
      
      glEnable(GL_BLEND); 
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
      glBindTexture(GL_TEXTURE_2D,m_pArrowTexture->nID);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glColor3f(1,1,1);
      glTexCoord2f(0,0);
      getParent()->glVertex((C+p1).x,(C+p1).y);
      glTexCoord2f(1,0);
      getParent()->glVertex((C+p2).x,(C+p2).y);
      glTexCoord2f(1,1);
      getParent()->glVertex((C+p3).x,(C+p3).y);
      glTexCoord2f(0,1);
      getParent()->glVertex((C+p4).x,(C+p4).y);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);        
    }
        
    /* Messages */
    if(pGame != NULL) {
      for(int i=0;i<pGame->getGameMessage().size();i++) {
        GameMessage *pMsg = pGame->getGameMessage()[i];
        int x1,y1,x2,y2;
        UITextDraw::getTextExt(getMediumFont(),pMsg->Text,&x1,&y1,&x2,&y2);
        UITextDraw::printRaw(getMediumFont(),400 - (x2-x1)/2,pMsg->Pos[1]*600,pMsg->Text,MAKE_COLOR(255,255,255,pMsg->nAlpha));
      } 
    }
  }
  
  /*===========================================================================
  Sprite rendering main
  ===========================================================================*/
  void GameRenderer::_RenderSprites(bool bForeground,bool bBackground) { 
    MotoGame *pGame = getGameObject();
  
    /* In front? */
    if(bForeground) {
      for(int i=0;i<pGame->getFSprites().size();i++)
        _RenderSprite(pGame->getFSprites()[i]);
    }

    /* Those in back? */
    if(bBackground) {
      for(int i=0;i<pGame->getBSprites().size();i++)
        _RenderSprite(pGame->getBSprites()[i]);
    }
  }

  /*===========================================================================
  Render a sprite
  ===========================================================================*/
  void GameRenderer::_RenderSprite(Entity *pSprite) {    
    /* TODO: make type fetching faster */
    SpriteType *pType = _GetSpriteTypeByName(pSprite->SpriteType);
    if(pType != NULL) {
      /* Get texture */
      GLuint GLName = pType->pTexture->nID;      
    
      /* Draw it */
      Vector2f p0,p1,p2,p3;
      
      p0 = Vector2f(pSprite->Pos.x,pSprite->Pos.y) +
           Vector2f(-pType->Center.x,-pType->Center.y);
      p1 = Vector2f(pSprite->Pos.x+pType->Size.x,pSprite->Pos.y) +
           Vector2f(-pType->Center.x,-pType->Center.y);
      p2 = Vector2f(pSprite->Pos.x+pType->Size.x,pSprite->Pos.y+pType->Size.y) +
           Vector2f(-pType->Center.x,-pType->Center.y);
      p3 = Vector2f(pSprite->Pos.x,pSprite->Pos.y+pType->Size.y) +
           Vector2f(-pType->Center.x,-pType->Center.y);
            
      glEnable(GL_BLEND); 
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
      glBindTexture(GL_TEXTURE_2D,GLName);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);   
      glColor3f(1,1,1);   
      glTexCoord2f(0,1);
      _Vertex(p0);
      glTexCoord2f(1,1);
      _Vertex(p1);
      glTexCoord2f(1,0);
      _Vertex(p2);
      glTexCoord2f(0,0);
      _Vertex(p3);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);

      /* Debug mode? */
      if(isDebug()) {
        /* Sprite marker */
        glBegin(GL_LINE_STRIP);      
        glColor3f(1,0.5,0.2);
        _Vertex(pSprite->Pos - Vector2f(0.5,0));
        _Vertex(pSprite->Pos + Vector2f(0.5,0));
        glEnd();
        glBegin(GL_LINE_STRIP);      
        glColor3f(1,0.5,0.2);
        _Vertex(pSprite->Pos - Vector2f(0,0.5));
        _Vertex(pSprite->Pos + Vector2f(0,0.5));
        glEnd();
        
        _DbgText(pSprite->Pos,pSprite->ID,MAKE_COLOR(255,128,51,255));
      }
    }
  }
       
  /*===========================================================================
  Rendering of the bike
  ===========================================================================*/
  void GameRenderer::_RenderBike(void) {
    /* Render bike */
    BikeState *pBike = getGameObject()->getBikeState();
    
    Vector2f p0,p1,p2,p3,o0,o1,o2,o3;
    Vector2f C;
    Vector2f Sv,Rc,Fc;
      
    /* Draw front wheel */
    o0 = Vector2f(-pBike->pParams->WR,pBike->pParams->WR);
    o1 = Vector2f(pBike->pParams->WR,pBike->pParams->WR);
    o2 = Vector2f(pBike->pParams->WR,-pBike->pParams->WR);
    o3 = Vector2f(-pBike->pParams->WR,-pBike->pParams->WR);
    p0 = Vector2f(o0.x*pBike->pfFrontWheelRot[0*4+0] + o0.y*pBike->pfFrontWheelRot[0*4+1],
                  o0.x*pBike->pfFrontWheelRot[1*4+0] + o0.y*pBike->pfFrontWheelRot[1*4+1]);
    p1 = Vector2f(o1.x*pBike->pfFrontWheelRot[0*4+0] + o1.y*pBike->pfFrontWheelRot[0*4+1],
                  o1.x*pBike->pfFrontWheelRot[1*4+0] + o1.y*pBike->pfFrontWheelRot[1*4+1]);
    p2 = Vector2f(o2.x*pBike->pfFrontWheelRot[0*4+0] + o2.y*pBike->pfFrontWheelRot[0*4+1],
                  o2.x*pBike->pfFrontWheelRot[1*4+0] + o2.y*pBike->pfFrontWheelRot[1*4+1]);
    p3 = Vector2f(o3.x*pBike->pfFrontWheelRot[0*4+0] + o3.y*pBike->pfFrontWheelRot[0*4+1],
                  o3.x*pBike->pfFrontWheelRot[1*4+0] + o3.y*pBike->pfFrontWheelRot[1*4+1]);
    
    C = Vector2f(pBike->pfFrontWheelPos[0],pBike->pfFrontWheelPos[1]);
    Fc = (p0 + p1 + p2 + p3) * 0.25f + C;
    
    glEnable(GL_BLEND); 
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
    glBindTexture(GL_TEXTURE_2D,m_pBikeWheel1->nID);
    glEnable(GL_TEXTURE_2D);
    glBegin(GL_POLYGON);
    glColor3f(1,1,1);
    glTexCoord2f(0,1);
    _Vertex(p0+C);    
    glTexCoord2f(1,1);
    _Vertex(p1+C);
    glTexCoord2f(1,0);
    _Vertex(p2+C);
    glTexCoord2f(0,0);
    _Vertex(p3+C);
    glEnd();
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);

    /* Draw rear wheel */
    o0 = Vector2f(-pBike->pParams->WR,pBike->pParams->WR);
    o1 = Vector2f(pBike->pParams->WR,pBike->pParams->WR);
    o2 = Vector2f(pBike->pParams->WR,-pBike->pParams->WR);
    o3 = Vector2f(-pBike->pParams->WR,-pBike->pParams->WR);
    p0 = Vector2f(o0.x*pBike->pfRearWheelRot[0*4+0] + o0.y*pBike->pfRearWheelRot[0*4+1],
                  o0.x*pBike->pfRearWheelRot[1*4+0] + o0.y*pBike->pfRearWheelRot[1*4+1]);
    p1 = Vector2f(o1.x*pBike->pfRearWheelRot[0*4+0] + o1.y*pBike->pfRearWheelRot[0*4+1],
                  o1.x*pBike->pfRearWheelRot[1*4+0] + o1.y*pBike->pfRearWheelRot[1*4+1]);
    p2 = Vector2f(o2.x*pBike->pfRearWheelRot[0*4+0] + o2.y*pBike->pfRearWheelRot[0*4+1],
                  o2.x*pBike->pfRearWheelRot[1*4+0] + o2.y*pBike->pfRearWheelRot[1*4+1]);
    p3 = Vector2f(o3.x*pBike->pfRearWheelRot[0*4+0] + o3.y*pBike->pfRearWheelRot[0*4+1],
                  o3.x*pBike->pfRearWheelRot[1*4+0] + o3.y*pBike->pfRearWheelRot[1*4+1]);
    
    C = Vector2f(pBike->pfRearWheelPos[0],pBike->pfRearWheelPos[1]);        
    Rc = (p0 + p1 + p2 + p3) * 0.25f + C;
    
    glEnable(GL_BLEND); 
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
    glBindTexture(GL_TEXTURE_2D,m_pBikeWheel1->nID);
    glEnable(GL_TEXTURE_2D);
    glBegin(GL_POLYGON);
    glColor3f(1,1,1);
    glTexCoord2f(0,1);
    _Vertex(p0+C);    
    glTexCoord2f(1,1);
    _Vertex(p1+C);
    glTexCoord2f(1,0);
    _Vertex(p2+C);
    glTexCoord2f(0,0);
    _Vertex(p3+C);
    glEnd();
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);

    /* Draw swing arm */
    if(pBike->Dir == DD_RIGHT) {       
      Sv = pBike->SwingAnchorP - Rc;
      Sv.normalize();
      p0 = pBike->RearWheelP + Vector2f(-Sv.y,Sv.x)*0.07f - Sv*0.08f;
      p1 = pBike->SwingAnchorP + Vector2f(-Sv.y,Sv.x)*0.07f;
      p2 = pBike->SwingAnchorP - Vector2f(-Sv.y,Sv.x)*0.07f;
      p3 = pBike->RearWheelP - Vector2f(-Sv.y,Sv.x)*0.07f - Sv*0.08f;
    }
    else {
      Sv = pBike->SwingAnchor2P - Fc;
      Sv.normalize();
      p0 = pBike->FrontWheelP + Vector2f(-Sv.y,Sv.x)*0.07f - Sv*0.08f;
      p1 = pBike->SwingAnchor2P + Vector2f(-Sv.y,Sv.x)*0.07f;
      p2 = pBike->SwingAnchor2P - Vector2f(-Sv.y,Sv.x)*0.07f;
      p3 = pBike->FrontWheelP - Vector2f(-Sv.y,Sv.x)*0.07f - Sv*0.08f;
    }        

    glEnable(GL_BLEND); 
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
    glBindTexture(GL_TEXTURE_2D,m_pBikeRear1->nID);
    glEnable(GL_TEXTURE_2D);
    glBegin(GL_POLYGON);
    glColor3f(1,1,1);
    glTexCoord2f(0,0);
    _Vertex(p0);    
    glTexCoord2f(1,0);
    _Vertex(p1);
    glTexCoord2f(1,1);
    _Vertex(p2);
    glTexCoord2f(0,1);
    _Vertex(p3);
    glEnd();
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);        

    /* Draw front suspension */
    if(pBike->Dir == DD_RIGHT) {
      Sv = pBike->FrontAnchorP - Fc;
      Sv.normalize();         
      p0 = pBike->FrontWheelP + Vector2f(-Sv.y,Sv.x)*0.04f - Sv*0.05f;
      p1 = pBike->FrontAnchorP + Vector2f(-Sv.y,Sv.x)*0.04f - Sv*0.35f;
      p2 = pBike->FrontAnchorP - Vector2f(-Sv.y,Sv.x)*0.04f - Sv*0.35f;
      p3 = pBike->FrontWheelP - Vector2f(-Sv.y,Sv.x)*0.04f - Sv*0.05f;
    }
    else {
      Sv = pBike->FrontAnchor2P - Rc;
      Sv.normalize();         
      p0 = pBike->RearWheelP + Vector2f(-Sv.y,Sv.x)*0.04f - Sv*0.05f;
      p1 = pBike->FrontAnchor2P + Vector2f(-Sv.y,Sv.x)*0.04f - Sv*0.35f;
      p2 = pBike->FrontAnchor2P - Vector2f(-Sv.y,Sv.x)*0.04f - Sv*0.35f;
      p3 = pBike->RearWheelP - Vector2f(-Sv.y,Sv.x)*0.04f - Sv*0.05f;
    }
    
    glEnable(GL_BLEND); 
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
    glBindTexture(GL_TEXTURE_2D,m_pBikeFront1->nID);
    glEnable(GL_TEXTURE_2D);
    glBegin(GL_POLYGON);
    glColor3f(1,1,1);
    glTexCoord2f(1,0);
    _Vertex(p0);    
    glTexCoord2f(1,1);
    _Vertex(p1);
    glTexCoord2f(0,1);
    _Vertex(p2);
    glTexCoord2f(0,0);
    _Vertex(p3);
    glEnd();
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);        

    /* Draw body/frame */
    o0 = Vector2f(-1,0.5);
    o1 = Vector2f(1,0.5);
    o2 = Vector2f(1,-0.5);
    o3 = Vector2f(-1,-0.5);
    p0 = Vector2f(o0.x*pBike->pfFrameRot[0*4+0] + o0.y*pBike->pfFrameRot[0*4+1],
                  o0.x*pBike->pfFrameRot[1*4+0] + o0.y*pBike->pfFrameRot[1*4+1]);
    p1 = Vector2f(o1.x*pBike->pfFrameRot[0*4+0] + o1.y*pBike->pfFrameRot[0*4+1],
                  o1.x*pBike->pfFrameRot[1*4+0] + o1.y*pBike->pfFrameRot[1*4+1]);
    p2 = Vector2f(o2.x*pBike->pfFrameRot[0*4+0] + o2.y*pBike->pfFrameRot[0*4+1],
                  o2.x*pBike->pfFrameRot[1*4+0] + o2.y*pBike->pfFrameRot[1*4+1]);
    p3 = Vector2f(o3.x*pBike->pfFrameRot[0*4+0] + o3.y*pBike->pfFrameRot[0*4+1],
                  o3.x*pBike->pfFrameRot[1*4+0] + o3.y*pBike->pfFrameRot[1*4+1]);
    
    C = Vector2f(pBike->pfFramePos[0],pBike->pfFramePos[1]);        
    
    glEnable(GL_BLEND); 
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
    glBindTexture(GL_TEXTURE_2D,m_pBikeBody1->nID);
    glEnable(GL_TEXTURE_2D);
    glBegin(GL_POLYGON);
    glColor3f(1,1,1);
    if(pBike->Dir == DD_RIGHT) {
      glTexCoord2f(0,0);
      _Vertex(p0+C);    
      glTexCoord2f(1,0);
      _Vertex(p1+C);
      glTexCoord2f(1,1);
      _Vertex(p2+C);
      glTexCoord2f(0,1);
      _Vertex(p3+C);
    }
    else {
      glTexCoord2f(1,0);
      _Vertex(p0+C);    
      glTexCoord2f(0,0);
      _Vertex(p1+C);
      glTexCoord2f(0,1);
      _Vertex(p2+C);
      glTexCoord2f(1,1);
      _Vertex(p3+C);
    }
    glEnd();
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    
    if(pBike->Dir == DD_RIGHT) {
      /* Draw rider torso */
      Sv = pBike->ShoulderP - pBike->LowerBodyP;
      Sv.normalize();         
      p0 = pBike->ShoulderP + Vector2f(-Sv.y,Sv.x)*0.24f + Sv*0.46f;
      p1 = pBike->LowerBodyP + Vector2f(-Sv.y,Sv.x)*0.24f - Sv*0.1f;
      p2 = pBike->LowerBodyP - Vector2f(-Sv.y,Sv.x)*0.24f - Sv*0.1f;
      p3 = pBike->ShoulderP - Vector2f(-Sv.y,Sv.x)*0.24f + Sv*0.46f;
      
      glEnable(GL_BLEND); 
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
      glBindTexture(GL_TEXTURE_2D,m_pRiderTorso1->nID);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glColor3f(1,1,1);
      glTexCoord2f(0,0);
      _Vertex(p0);    
      glTexCoord2f(0,1);
      _Vertex(p1);
      glTexCoord2f(1,1);
      _Vertex(p2);
      glTexCoord2f(1,0);
      _Vertex(p3);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);        
      
      /* Draw rider upper arm */
      Sv = pBike->ShoulderP - pBike->ElbowP;
      Sv.normalize();         
      p0 = pBike->ShoulderP + Vector2f(-Sv.y,Sv.x)*0.12f;
      p1 = pBike->ElbowP + Vector2f(-Sv.y,Sv.x)*0.12f - Sv*0.05f;
      p2 = pBike->ElbowP - Vector2f(-Sv.y,Sv.x)*0.10f - Sv*0.05f;
      p3 = pBike->ShoulderP - Vector2f(-Sv.y,Sv.x)*0.10f;
      
      glEnable(GL_BLEND); 
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
      glBindTexture(GL_TEXTURE_2D,m_pRiderUpperArm1->nID);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glColor3f(1,1,1);
      glTexCoord2f(0,0);
      _Vertex(p0);    
      glTexCoord2f(0,1);
      _Vertex(p1);
      glTexCoord2f(1,1);
      _Vertex(p2);
      glTexCoord2f(1,0);
      _Vertex(p3);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);        

      /* Draw rider lower arm */
      Sv = pBike->ElbowP - pBike->HandP;
      Sv.normalize();         
      p0 = pBike->ElbowP + Vector2f(-Sv.y,Sv.x)*0.12f + Sv*0.09f;
      p1 = pBike->HandP + Vector2f(-Sv.y,Sv.x)*0.12f - Sv*0.05f;
      p2 = pBike->HandP - Vector2f(-Sv.y,Sv.x)*0.10f - Sv*0.05f;
      p3 = pBike->ElbowP - Vector2f(-Sv.y,Sv.x)*0.10f + Sv*0.09f;
      
      glEnable(GL_BLEND); 
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
      glBindTexture(GL_TEXTURE_2D,m_pRiderLowerArm1->nID);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glColor3f(1,1,1);
      glTexCoord2f(0,0);
      _Vertex(p0);    
      glTexCoord2f(1,0);
      _Vertex(p1);
      glTexCoord2f(1,1);
      _Vertex(p2);
      glTexCoord2f(0,1);
      _Vertex(p3);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);        
      

      /* Draw rider upper leg */
      Sv = pBike->LowerBodyP - pBike->KneeP;
      Sv.normalize();         
      p0 = pBike->LowerBodyP + Vector2f(-Sv.y,Sv.x)*0.20f + Sv*0.14f;
      p1 = pBike->KneeP + Vector2f(-Sv.y,Sv.x)*0.15f + Sv*0.0f;
      p2 = pBike->KneeP - Vector2f(-Sv.y,Sv.x)*0.15f + Sv*0.0f;
      p3 = pBike->LowerBodyP - Vector2f(-Sv.y,Sv.x)*0.1f + Sv*0.14f;
      
      glEnable(GL_BLEND); 
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
      glBindTexture(GL_TEXTURE_2D,m_pRiderUpperLeg1->nID);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glColor3f(1,1,1);
      glTexCoord2f(0,1);
      _Vertex(p0);    
      glTexCoord2f(1,1);
      _Vertex(p1);
      glTexCoord2f(1,0);
      _Vertex(p2);
      glTexCoord2f(0,0);
      _Vertex(p3);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);        

      /* Draw rider lower leg */
      Sv = pBike->KneeP - pBike->FootP;
      Sv.normalize();         
      p0 = pBike->KneeP + Vector2f(-Sv.y,Sv.x)*0.23f + Sv*0.01f;
      p1 = pBike->FootP + Vector2f(-Sv.y,Sv.x)*0.2f;
      p2 = pBike->FootP - Vector2f(-Sv.y,Sv.x)*0.2f;
      p3 = pBike->KneeP - Vector2f(-Sv.y,Sv.x)*0.23f + Sv*0.1f;
      
      glEnable(GL_BLEND); 
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
      glBindTexture(GL_TEXTURE_2D,m_pRiderLowerLeg1->nID);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glColor3f(1,1,1);
      glTexCoord2f(0,0);
      _Vertex(p0);    
      glTexCoord2f(0,1);
      _Vertex(p1);
      glTexCoord2f(1,1);
      _Vertex(p2);
      glTexCoord2f(1,0);
      _Vertex(p3);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);        
    }
    else if(pBike->Dir == DD_LEFT) {
      /* Draw rider torso */
      Sv = pBike->Shoulder2P - pBike->LowerBody2P;
      Sv.normalize();         
      p0 = pBike->Shoulder2P + Vector2f(-Sv.y,Sv.x)*0.24f + Sv*0.46f;
      p1 = pBike->LowerBody2P + Vector2f(-Sv.y,Sv.x)*0.24f - Sv*0.1f;
      p2 = pBike->LowerBody2P - Vector2f(-Sv.y,Sv.x)*0.24f - Sv*0.1f;
      p3 = pBike->Shoulder2P - Vector2f(-Sv.y,Sv.x)*0.24f + Sv*0.46f;
      
      glEnable(GL_BLEND); 
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
      glBindTexture(GL_TEXTURE_2D,m_pRiderTorso1->nID);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glColor3f(1,1,1);
      glTexCoord2f(1,0);
      _Vertex(p0);    
      glTexCoord2f(1,1);
      _Vertex(p1);
      glTexCoord2f(0,1);
      _Vertex(p2);
      glTexCoord2f(0,0);
      _Vertex(p3);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);        
      
      /* Draw rider upper arm */
      Sv = pBike->Shoulder2P - pBike->Elbow2P;
      Sv.normalize();         
      p0 = pBike->Shoulder2P + Vector2f(-Sv.y,Sv.x)*0.12f;
      p1 = pBike->Elbow2P + Vector2f(-Sv.y,Sv.x)*0.12f - Sv*0.05f;
      p2 = pBike->Elbow2P - Vector2f(-Sv.y,Sv.x)*0.10f - Sv*0.05f;
      p3 = pBike->Shoulder2P - Vector2f(-Sv.y,Sv.x)*0.10f;
      
      glEnable(GL_BLEND); 
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
      glBindTexture(GL_TEXTURE_2D,m_pRiderUpperArm1->nID);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glColor3f(1,1,1);
      glTexCoord2f(1,0);
      _Vertex(p0);    
      glTexCoord2f(1,1);
      _Vertex(p1);
      glTexCoord2f(0,1);
      _Vertex(p2);
      glTexCoord2f(0,0);
      _Vertex(p3);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);        

      /* Draw rider lower arm */
      Sv = pBike->Elbow2P - pBike->Hand2P;
      Sv.normalize();         
      p0 = pBike->Elbow2P + Vector2f(-Sv.y,Sv.x)*0.12f + Sv*0.09f;
      p1 = pBike->Hand2P + Vector2f(-Sv.y,Sv.x)*0.12f - Sv*0.05f;
      p2 = pBike->Hand2P - Vector2f(-Sv.y,Sv.x)*0.10f - Sv*0.05f;
      p3 = pBike->Elbow2P - Vector2f(-Sv.y,Sv.x)*0.10f + Sv*0.09f;
      
      glEnable(GL_BLEND); 
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
      glBindTexture(GL_TEXTURE_2D,m_pRiderLowerArm1->nID);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glColor3f(1,1,1);
      glTexCoord2f(0,1);
      _Vertex(p0);    
      glTexCoord2f(1,1);
      _Vertex(p1);
      glTexCoord2f(1,0);
      _Vertex(p2);
      glTexCoord2f(0,0);
      _Vertex(p3);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);              

      /* Draw rider upper leg */
      Sv = pBike->LowerBody2P - pBike->Knee2P;
      Sv.normalize();         
      p0 = pBike->LowerBody2P + Vector2f(-Sv.y,Sv.x)*0.20f + Sv*0.14f;
      p1 = pBike->Knee2P + Vector2f(-Sv.y,Sv.x)*0.15f + Sv*0.0f;
      p2 = pBike->Knee2P - Vector2f(-Sv.y,Sv.x)*0.15f + Sv*0.0f;
      p3 = pBike->LowerBody2P - Vector2f(-Sv.y,Sv.x)*0.1f + Sv*0.14f;
      
      glEnable(GL_BLEND); 
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
      glBindTexture(GL_TEXTURE_2D,m_pRiderUpperLeg1->nID);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glColor3f(1,1,1);
      glTexCoord2f(0,0);
      _Vertex(p0);    
      glTexCoord2f(1,0);
      _Vertex(p1);
      glTexCoord2f(1,1);
      _Vertex(p2);
      glTexCoord2f(0,1);
      _Vertex(p3);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);        

      /* Draw rider lower leg */
      Sv = pBike->Knee2P - pBike->Foot2P;
      Sv.normalize();         
      p0 = pBike->Knee2P + Vector2f(-Sv.y,Sv.x)*0.23f + Sv*0.01f;
      p1 = pBike->Foot2P + Vector2f(-Sv.y,Sv.x)*0.2f;
      p2 = pBike->Foot2P - Vector2f(-Sv.y,Sv.x)*0.2f;
      p3 = pBike->Knee2P - Vector2f(-Sv.y,Sv.x)*0.23f + Sv*0.1f;
      
      glEnable(GL_BLEND); 
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
      glBindTexture(GL_TEXTURE_2D,m_pRiderLowerLeg1->nID);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glColor3f(1,1,1);
      glTexCoord2f(1,0);
      _Vertex(p0);    
      glTexCoord2f(1,1);
      _Vertex(p1);
      glTexCoord2f(0,1);
      _Vertex(p2);
      glTexCoord2f(0,0);
      _Vertex(p3);
      glEnd();
      glDisable(GL_TEXTURE_2D);
      glDisable(GL_BLEND);     
    }   
  }

  void GameRenderer::_DrawRotatedMarker(Vector2f Pos,dReal *pfRot) {
    Vector2f p0,p1,p2,p3,o0,o1,o2,o3;
    Vector2f C = Pos;

    o0 = Vector2f(-0.1,0.1);
    o1 = Vector2f(0.1,0.1);
    o2 = Vector2f(0.1,-0.1);
    o3 = Vector2f(-0.1,-0.1);
    
    if(pfRot != NULL) {
      p0 = Vector2f(o0.x*pfRot[0*4+0] + o0.y*pfRot[0*4+1],
                    o0.x*pfRot[1*4+0] + o0.y*pfRot[1*4+1]);
      p1 = Vector2f(o1.x*pfRot[0*4+0] + o1.y*pfRot[0*4+1],
                    o1.x*pfRot[1*4+0] + o1.y*pfRot[1*4+1]);
      p2 = Vector2f(o2.x*pfRot[0*4+0] + o2.y*pfRot[0*4+1],
                    o2.x*pfRot[1*4+0] + o2.y*pfRot[1*4+1]);
      p3 = Vector2f(o3.x*pfRot[0*4+0] + o3.y*pfRot[0*4+1],
                    o3.x*pfRot[1*4+0] + o3.y*pfRot[1*4+1]);
    }
    else {
      p0 = o0;
      p1 = o1;
      p2 = o2;
      p3 = o3;
    }

    glBegin(GL_LINE_STRIP);
    glColor3f(1,1,1);
    _Vertex(p0+C);    
    _Vertex(p2+C);
    glEnd();
    glBegin(GL_LINE_STRIP);
    glColor3f(1,1,1);
    _Vertex(p1+C);    
    _Vertex(p3+C);
    glEnd();
  }
  
  /*===========================================================================
  Blocks
  ===========================================================================*/
  void GameRenderer::_RenderBlocks(void) {
    MotoGame *pGame = getGameObject();

    /* Render all non-background blocks */
    std::vector<ConvexBlock *> &Blocks = pGame->getBlocks();

    for(int i=0;i<Blocks.size();i++) {
      Vector2f Center;
      Texture *pTexture;
      GLuint GLName = 0;
      
      if(Blocks[i]->pSrcBlock && Blocks[i]->pSrcBlock->bBackground) continue;

      /* Main body */      
      if(Blocks[i]->pSrcBlock) {
        Center = Vector2f(Blocks[i]->pSrcBlock->fPosX,Blocks[i]->pSrcBlock->fPosY);
        pTexture = getParent()->TexMan.getTexture(Blocks[i]->pSrcBlock->Texture);
        if(pTexture != NULL) {GLName = pTexture->nID;}
      }
      else {
        pTexture = getParent()->TexMan.getTexture("default");
        if(pTexture != NULL) {GLName = pTexture->nID;}
      }
    
      glBindTexture(GL_TEXTURE_2D,GLName);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glBegin(GL_LINE_LOOP);      
      glColor3f(1,1,1);
      for(int j=0;j<Blocks[i]->Vertices.size();j++) {
        glTexCoord2f((Center.x+Blocks[i]->Vertices[j]->P.x) * 0.25,
                      (Center.y+Blocks[i]->Vertices[j]->P.y) * 0.25);
        _Vertex( Center + Blocks[i]->Vertices[j]->P );                  
      }
      glEnd();
      glDisable(GL_TEXTURE_2D);            
    }    
    
    /* Render all special edges */
    for(int i=0;i<pGame->getOverlayEdges().size();i++) {
      OverlayEdge *pEdge = pGame->getOverlayEdges()[i];
      
      switch(pEdge->Effect) {
        case EE_GRASS:
          if(m_pEdgeGrass1 != NULL) {
            glEnable(GL_BLEND); 
            glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
            glBindTexture(GL_TEXTURE_2D,m_pEdgeGrass1->nID);
            glEnable(GL_TEXTURE_2D);
            glBegin(GL_POLYGON);
            glColor3f(1,1,1);
            glTexCoord2f((pEdge->pSrcBlock->fPosX+pEdge->P1.x)*0.5,0.01);
            _Vertex(pEdge->P1 + Vector2f(pEdge->pSrcBlock->fPosX,pEdge->pSrcBlock->fPosY));
            glTexCoord2f((pEdge->pSrcBlock->fPosX+pEdge->P2.x)*0.5,0.01);
            _Vertex(pEdge->P2 + Vector2f(pEdge->pSrcBlock->fPosX,pEdge->pSrcBlock->fPosY));
            glTexCoord2f((pEdge->pSrcBlock->fPosX+pEdge->P2.x)*0.5,0.99);
            _Vertex(pEdge->P2 + Vector2f(pEdge->pSrcBlock->fPosX,pEdge->pSrcBlock->fPosY) + Vector2f(0,-0.3));
            glTexCoord2f((pEdge->pSrcBlock->fPosX+pEdge->P1.x)*0.5,0.99);
            _Vertex(pEdge->P1 + Vector2f(pEdge->pSrcBlock->fPosX,pEdge->pSrcBlock->fPosY) + Vector2f(0,-0.3));
            glEnd();
            glDisable(GL_TEXTURE_2D);
            glDisable(GL_BLEND);
          }
          break;
      }
    }
  }  
  
  /*===========================================================================
  And background rendering
  ===========================================================================*/
  void GameRenderer::_RenderBackground(void) { 
    MotoGame *pGame = getGameObject();

    /* Render sky */
    if(m_pSkyTexture1 != NULL) {
      glBindTexture(GL_TEXTURE_2D,m_pSkyTexture1->nID);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);   
      glColor3f(1,1,1);   
      glTexCoord2f(-m_Scroll.x*0.015,m_Scroll.y*0.015);
      getParent()->glVertex(0,0);
      glTexCoord2f(-m_Scroll.x*0.015+0.5,m_Scroll.y*0.015);
      getParent()->glVertex(getParent()->getDispWidth(),0);
      glTexCoord2f(-m_Scroll.x*0.015+0.5,m_Scroll.y*0.015+0.5);
      getParent()->glVertex(getParent()->getDispWidth(),getParent()->getDispHeight());
      glTexCoord2f(-m_Scroll.x*0.015,m_Scroll.y*0.015+0.5);
      getParent()->glVertex(0,getParent()->getDispHeight());
      glEnd();
      glDisable(GL_TEXTURE_2D); 
    }   
    
    /* Render background blocks */
    std::vector<ConvexBlock *> &Blocks = pGame->getBlocks();

    for(int i=0;i<Blocks.size();i++) {
      Vector2f Center;
      Texture *pTexture;
      GLuint GLName = 0;
      
      if(!Blocks[i]->pSrcBlock || !Blocks[i]->pSrcBlock->bBackground) continue;

      /* Main body */      
      Center = Vector2f(Blocks[i]->pSrcBlock->fPosX,Blocks[i]->pSrcBlock->fPosY);
      pTexture = getParent()->TexMan.getTexture(Blocks[i]->pSrcBlock->Texture);
      if(pTexture != NULL) {GLName = pTexture->nID;}
    
      glBindTexture(GL_TEXTURE_2D,GLName);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_POLYGON);
      glColor3f(1,1,1);
      for(int j=0;j<Blocks[i]->Vertices.size();j++) {
        glTexCoord2f((Center.x+Blocks[i]->Vertices[j]->P.x) * 0.25,
                      (Center.y+Blocks[i]->Vertices[j]->P.y) * 0.25);
        _Vertex( Center + Blocks[i]->Vertices[j]->P );                  
      }
      glEnd();
      glDisable(GL_TEXTURE_2D);
    }        
  }

  /*===========================================================================
  And general entity rendering
  ===========================================================================*/
  void GameRenderer::_RenderEntities(void) {
    MotoGame *pGame = getGameObject();

    /* Render all entities */
    std::vector<Entity *> &Entities = pGame->getEntities();    
        
    for(int i=0;i<Entities.size();i++) {
      Entity *pEntity = Entities[i];
      
      if(pEntity->Type == ET_STRAWBERRY) {
        /* Draw strawberry */
        _DrawAnimation(pEntity->Pos,m_pStrawberryAnim);
      }
      else if(pEntity->Type == ET_ENDOFLEVEL) {
        /* Draw end-of-level flower */
        _DrawAnimation(pEntity->Pos,m_pFlowerAnim);
      }
      else if(pEntity->Type == ET_WRECKER) {
        /* Draw nasty wrecker */
        _DrawAnimation(pEntity->Pos,m_pWreckerAnim);
      }
      
      /* If this is debug-mode, also draw entity's area of effect */
      if(isDebug()) {
        Vector2f C = Vector2f( getParent()->getDispWidth()/2 + (float)(pEntity->Pos.x + m_Scroll.x)*m_fZoom,
                               getParent()->getDispHeight()/2 - (float)(pEntity->Pos.y + m_Scroll.y)*m_fZoom );
        Vector2f Cr = Vector2f( getParent()->getDispWidth()/2 + (float)((pEntity->Pos.x + pEntity->fSize) + m_Scroll.x)*m_fZoom,
                                getParent()->getDispHeight()/2 - (float)(pEntity->Pos.y + m_Scroll.y)*m_fZoom );
        float r = (C-Cr).length();
        getParent()->drawCircle(C,r,1,0,MAKE_COLOR(0,255,0,255));
      }
    }
  }

  /*===========================================================================
  Helpers
  ===========================================================================*/
  void GameRenderer::_Vertex(Vector2f P) {
    getParent()->glVertex( getParent()->getDispWidth()/2 + (float)(P.x + m_Scroll.x)*m_fZoom,
                           getParent()->getDispHeight()/2 - (float)(P.y + m_Scroll.y)*m_fZoom );
  }
  
  SpriteType *GameRenderer::_GetSpriteTypeByName(std::string Name) {
    for(int i=0;i<m_nNumSpriteTypes;i++)
      if(m_SpriteTypes[i].Name == Name) return &m_SpriteTypes[i];
    return NULL;
  }
  
  void GameRenderer::_DbgText(Vector2f P,std::string Text,Color c) {
    Vector2f Sp = Vector2f(getParent()->getDispWidth()/2 + (float)(P.x + m_Scroll.x)*m_fZoom,
                           getParent()->getDispHeight()/2 - (float)(P.y + m_Scroll.y)*m_fZoom) -
                  Vector2f(getParent()->getTextWidth(Text)/2.0f,getParent()->getTextHeight(Text)/2.0f);
    getParent()->drawText(Sp,Text,0,c,true);
  }   
  
  /*===========================================================================
  Animations
  ===========================================================================*/
  Animation *GameRenderer::_GetAnimationByName(std::string Name) {
    for(int i=0;i<m_Anims.size();i++)
      if(m_Anims[i]->Name == Name) return m_Anims[i];
    return NULL;
  }
  
  void GameRenderer::_UpdateAnimations(void) {
    for(int i=0;i<m_Anims.size();i++) {
      Animation *pAnim = m_Anims[i];
    
      /* Next frame? */
      if(getParent()->getRealTime() > pAnim->fFrameTime+pAnim->m_Frames[pAnim->m_nCurFrame]->fDelay) {
        pAnim->fFrameTime=getParent()->getRealTime();
        pAnim->m_nCurFrame++;
        if(pAnim->m_nCurFrame == pAnim->m_Frames.size()) pAnim->m_nCurFrame = 0;      
      }
    }
  }    

  void GameRenderer::_DrawAnimation(Vector2f Pos,Animation *pAnim) {
    AnimationFrame *pFrame = pAnim->m_Frames[pAnim->m_nCurFrame];
    
    /* Get texture */
    GLuint GLName = pFrame->pTexture->nID;      
  
    /* Draw it */
    Vector2f p0,p1,p2,p3;
    
    p0 = Vector2f(Pos.x,Pos.y) +
          Vector2f(-pFrame->Center.x,-pFrame->Center.y);
    p1 = Vector2f(Pos.x+pFrame->Size.x,Pos.y) +
          Vector2f(-pFrame->Center.x,-pFrame->Center.y);
    p2 = Vector2f(Pos.x+pFrame->Size.x,Pos.y+pFrame->Size.y) +
          Vector2f(-pFrame->Center.x,-pFrame->Center.y);
    p3 = Vector2f(Pos.x,Pos.y+pFrame->Size.y) +
          Vector2f(-pFrame->Center.x,-pFrame->Center.y);
          
    glEnable(GL_BLEND); 
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
    glBindTexture(GL_TEXTURE_2D,GLName);
    glEnable(GL_TEXTURE_2D);
    glBegin(GL_POLYGON);   
    glColor3f(1,1,1);   
    glTexCoord2f(0,1);
    _Vertex(p0);
    glTexCoord2f(1,1);
    _Vertex(p1);
    glTexCoord2f(1,0);
    _Vertex(p2);
    glTexCoord2f(0,0);
    _Vertex(p3);
    glEnd();
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);    
  }

  /*===========================================================================
  Free stuff
  ===========================================================================*/
  void GameRenderer::_Free(void) {
    /* Free animations */
    for(int i=0;i<m_Anims.size();i++) { 
      for(int j=0;j<m_Anims[i]->m_Frames.size();j++) 
        delete m_Anims[i]->m_Frames[j];
      delete m_Anims[i];
    }
    
    /* Free any particles left */
    for(int i=0;i<m_Particles.size();i++)
      delete m_Particles[i];
  }

  /*===========================================================================
  Debug info. Note how this is leaked into the void and nobody cares :) 
  ===========================================================================*/
  void GameRenderer::_RenderDebugInfo(void) {
    for(int i=0;i<m_DebugInfo.size();i++) {
      if(m_DebugInfo[i]->Type == "@WHITEPOLYGONS") {
        glBegin(GL_LINE_LOOP);
        glColor3f(1,1,1);
        for(int j=0;j<m_DebugInfo[i]->Args.size()/2;j++) {
          float x = atof(m_DebugInfo[i]->Args[j*2].c_str());
          float y = atof(m_DebugInfo[i]->Args[j*2+1].c_str());
          getParent()->glVertex(400 + x*10,300 - y*10);
        }
        glEnd();
      }
      else if(m_DebugInfo[i]->Type == "@REDLINES") {
        glBegin(GL_LINE_STRIP);
        glColor3f(1,0,0);
        for(int j=0;j<m_DebugInfo[i]->Args.size()/2;j++) {
          float x = atof(m_DebugInfo[i]->Args[j*2].c_str());
          float y = atof(m_DebugInfo[i]->Args[j*2+1].c_str());
          getParent()->glVertex(400 + x*10,300 - y*10);
        }
        glEnd();
      }
      else if(m_DebugInfo[i]->Type == "@GREENLINES") {
        glBegin(GL_LINE_STRIP);
        glColor3f(0,1,0);
        for(int j=0;j<m_DebugInfo[i]->Args.size()/2;j++) {
          float x = atof(m_DebugInfo[i]->Args[j*2].c_str());
          float y = atof(m_DebugInfo[i]->Args[j*2+1].c_str());
          getParent()->glVertex(400 + x*10,300 - y*10);
        }
        glEnd();
      }
    }
  }
  
  void GameRenderer::loadDebugInfo(std::string File) {
    FileHandle *pfh = FS::openIFile(File);
    if(pfh != NULL) {
      std::string Line;
      std::string Type = "";
      while(FS::readNextLine(pfh,Line)) {
        if(!Line.empty() && Line[0] != '#') {
          if(Line[0] == '@') {
            Type = Line;
          }
          else {
            GraphDebugInfo *p = new GraphDebugInfo;
            m_DebugInfo.push_back(p);
            p->Type = Type;
            while(1) {
              int k = Line.find_first_of(" ");
              if(k<0) {
                p->Args.push_back(Line);
                break;
              }
              else {
                std::string s = Line.substr(0,k);
                Line = Line.substr(k+1,Line.length()-k);
                p->Args.push_back(s);
              }
            }
          }
        }
      }
      FS::closeFile(pfh);
    }
  }

  /*===========================================================================
  Particles, rendering and management
  ===========================================================================*/
  Particle *GameRenderer::spawnParticle(ParticleType Type,Vector2f Pos,Vector2f Vel,float fLifeTime) {
    MotoGame *pGame = getGameObject();

    if(pGame != NULL) {
      Particle *p = new Particle;
      p->Type = Type;
      p->Pos = Pos;
      p->Vel = Vel;     
      p->fSpawnTime = pGame->getTime();
      p->fKillTime = pGame->getTime() + fLifeTime;
      p->Acc = Vector2f(0,0);
      p->fAng = 0;
      p->fAngAcc = 0;
      p->fAngVel = 0;
      m_Particles.push_back(p);
      
      int cc;
      switch(Type) {
        case PT_SMOKE1:
        case PT_SMOKE2:
          cc = (int)randomNum(0,50);
          p->SmokeColor = MAKE_COLOR(cc,cc,cc,255);
          p->fSmokeSize = randomNum(0,0.2);
          p->fAngVel = randomNum(-60,60);
          p->fKillTime += 10000;
          break;
        
        case PT_FIRE:
          p->FireColor = MAKE_COLOR(255,255,200,255);
          p->fFireSize = 0.1f;
          p->fKillTime += 10000;
          p->fFireSeed = randomNum(0,100);
          break;
      }
      
      return p;
    }
    
    return NULL;
  }  

  void GameRenderer::_UpdateParticles(float fTimeStep) {
    MotoGame *pGame = getGameObject();

    if(pGame != NULL) {
      /* Look for particle sources */
      {
        for(int i=0;i<pGame->getEntities().size();i++) {
          if(pGame->getEntities()[i]->Type == ET_PARTICLESOURCE && 
            pGame->getTime()>pGame->getEntities()[i]->fNextParticleTime) {
            
            if(pGame->getEntities()[i]->ParticleType == "Smoke") {
              /* Generate smoke */
              if(randomNum(0,1) < 0.5)
                spawnParticle(PT_SMOKE1,pGame->getEntities()[i]->Pos,Vector2f(randomNum(-0.6,0.6),randomNum(0.2,0.6)),0);
              else
                spawnParticle(PT_SMOKE2,pGame->getEntities()[i]->Pos,Vector2f(randomNum(-0.6,0.6),randomNum(0.2,0.6)),0);
              pGame->getEntities()[i]->fNextParticleTime = pGame->getTime() + randomNum(0.2,0.5f);
            }
            else if(pGame->getEntities()[i]->ParticleType == "Fire") {
              for(int k=0;k<10;k++) {
                /* Generate fire */
                spawnParticle(PT_FIRE,pGame->getEntities()[i]->Pos,Vector2f(randomNum(-1,1),randomNum(0.1,0.3)),0);
                pGame->getEntities()[i]->fNextParticleTime = pGame->getTime() + randomNum(0.05,0.1f);              
              }
            }
          }
        }
      }
    
      /* For each particle */
      int i=0;
      while(1) {
        if(i >= m_Particles.size()) break;
        bool bKillParticle = false;
        
        /* Do general handling of it */
        m_Particles[i]->Vel += m_Particles[i]->Acc * fTimeStep;
        m_Particles[i]->Pos += m_Particles[i]->Vel * fTimeStep;
        m_Particles[i]->fAngVel += m_Particles[i]->fAngAcc * fTimeStep;
        m_Particles[i]->fAng += m_Particles[i]->fAngVel * fTimeStep;
        
        if(pGame->getTime() > m_Particles[i]->fKillTime)
          bKillParticle = true;
        
        /* Do type dependant handling of it */
        float a,c,c2;
        
        switch(m_Particles[i]->Type) {
          case PT_SMOKE1:
          case PT_SMOKE2:
            m_Particles[i]->fSmokeSize += fTimeStep * 1.3f; /* grow */
            m_Particles[i]->Acc = Vector2f(0.2,0.5);  /* accelerate upwards */
            c=GET_RED(m_Particles[i]->SmokeColor);
            c += randomNum(40,50) * fTimeStep;
            a=GET_ALPHA(m_Particles[i]->SmokeColor);
            a -= 60.0f * fTimeStep;
            if(c>255) c=255;
            if(a<0) { 
              a=0;                       
              bKillParticle = true;
            }
            else
              m_Particles[i]->SmokeColor = MAKE_COLOR((int)c,(int)c,(int)c,(int)a);
            break;
            
          case PT_FIRE:
            c=GET_GREEN(m_Particles[i]->FireColor);
            c -= randomNum(190,210) * fTimeStep;
            if(c<0) c=0;
            c2=GET_BLUE(m_Particles[i]->FireColor);
            c2 -= randomNum(400,400) * fTimeStep;
            if(c2<0) c2=0;
            a=GET_ALPHA(m_Particles[i]->FireColor);
            a -= 180.0f * fTimeStep;
            if(a<0) { 
              a=0;                       
              bKillParticle = true;
            }
            else
              m_Particles[i]->FireColor = MAKE_COLOR(255,(int)c,(int)c2,(int)a);
            
            m_Particles[i]->Vel.x = sin((pGame->getTime() - 
                                         m_Particles[i]->fSpawnTime + m_Particles[i]->fFireSeed)*randomNum(5,15)) * 0.8f
                                      +
                                    sin((pGame->getTime()-m_Particles[i]->fFireSeed) * 10) * 0.3;
            m_Particles[i]->Acc.y = 2;
            break;
        }
        
        /* Show we kill it? */
        if(bKillParticle) {
          delete m_Particles[i];
          m_Particles.erase(m_Particles.begin() + i);
        }
        else i++;
      }
    }
  }  
  
  void GameRenderer::_RenderParticle(Vector2f P,Texture *pTexture,float fSize,float fAngle,Color c) {
    /* Render single particle */
    if(pTexture == NULL) return;
    
    Vector2f C = P;
    Vector2f p1,p2,p3,p4;
    p1 = Vector2f(1,0); p1.rotateXY(fAngle);
    p2 = Vector2f(1,0); p2.rotateXY(90+fAngle);
    p3 = Vector2f(1,0); p3.rotateXY(180+fAngle);
    p4 = Vector2f(1,0); p4.rotateXY(270+fAngle);
    
    p1 = C + p1 * fSize;
    p2 = C + p2 * fSize;
    p3 = C + p3 * fSize;
    p4 = C + p4 * fSize;
    
    glEnable(GL_BLEND); 
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);         
    glBindTexture(GL_TEXTURE_2D,pTexture->nID);
    glEnable(GL_TEXTURE_2D);
    glBegin(GL_POLYGON);
    glColor4ub(GET_RED(c),GET_GREEN(c),GET_BLUE(c),GET_ALPHA(c));
    glTexCoord2f(0,0);
    _Vertex(p1);
    glTexCoord2f(1,0);
    _Vertex(p2);
    glTexCoord2f(1,1);
    _Vertex(p3);
    glTexCoord2f(0,1);
    _Vertex(p4);
    glEnd();
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);            
  }
  
  void GameRenderer::_RenderParticles(void) {
    /* Render all particles */
    for(int i=0;i<m_Particles.size();i++) {
      switch(m_Particles[i]->Type) {
        case PT_SMOKE1:
          _RenderParticle(m_Particles[i]->Pos,m_pSmoke1,m_Particles[i]->fSmokeSize,
                          m_Particles[i]->fAng,m_Particles[i]->SmokeColor);
          break;
        case PT_SMOKE2:
          _RenderParticle(m_Particles[i]->Pos,m_pSmoke2,m_Particles[i]->fSmokeSize,
                          m_Particles[i]->fAng,m_Particles[i]->SmokeColor);
          break;
        case PT_FIRE:
          _RenderParticle(m_Particles[i]->Pos,m_pFire1,m_Particles[i]->fFireSize,
                          m_Particles[i]->fAng,m_Particles[i]->FireColor);
          break;
      }
    }
  }
  
};

